#coding: utf-8
"""
Google Maps 矩形区域 图片导出工具

ver 20130501

author smite

版权所有，请勿用于商业用途
"""
from globalmaptiles import *
import os
import threading
import urllib2
import pdb
import Image
import ImageDraw

class RectCreater():
    """
    经过算法确认，得到Google Maps瓦片坐标列表
    """
    def __init__(self,start_pos,end_pos,z_level):
        """
        start_pos=(x,y)
        end_pos=(x,y)
        
        start_pos--------------------------------
        -               area
        ---------------------------------------end_pos
        """
        self.start_pos=start_pos
        self.end_pos=end_pos
            
        self.z_level=z_level
        self.start_xyz=None
        self.end_xyz=None
        self.xyz_list=[]
        self.mercator = GlobalMercator()
        
        self.gen_start_end()
        self.gen_rect_list()
        
    def gen_start_end(self):
        """
        通过起始位置转换器瓦片坐标起始位置
        """
        mx1, my1 = self.mercator.LatLonToMeters( self.start_pos[0],self.start_pos[1] )
        mx2, my2 = self.mercator.LatLonToMeters( self.end_pos[0],self.end_pos[1] )
        tx1, ty1 = self.mercator.MetersToTile( mx1, my1,self.z_level )
        tx2, ty2 = self.mercator.MetersToTile( mx2, my2,self.z_level )
        self.start_xyz= self.mercator.GoogleTile(tx1, ty1, self.z_level)
        self.end_xyz= self.mercator.GoogleTile(tx2, ty2, self.z_level)
        
    def gen_rect_list(self):
        """
        循环取值
        """
        for y in range(self.start_xyz[1],self.end_xyz[1]+1):
            rect_x=[]
            for x in range(self.start_xyz[0],self.end_xyz[0]+1):
                rect_x.append((x,y,self.z_level))
                
            self.xyz_list.append(rect_x)
            
class ResourceGeter():
    """
    从RectCreater中获取瓦片坐标，下载对应图片到本地等待拼接
    """
    def __init__(self,xyz_list,dir_path,re_string,prefix='gm'):
        """
        xyz_list为list
        dir_path图片保存的文件夹
        re_string Google Maps服务器格式化字符串http://khms1.google.com/kh/v=128&src=app&x=%(x)s&y=%(y)s&z=%(z)s
        prefix保存文件的前缀
        """
        self.prefix=prefix
        self.xyz_list=xyz_list
        self.dir_path=dir_path
        self.web_404_dic={}
        self.cache_pics={}
        
        if os.path.exists(self.dir_path)==False:
            raise Exception('Save path not exists')
            
        self.re_string=re_string
        if len(self.re_string.split('%'))!=4:
            raise Exception('Google Maps Tile URL Error')
            
        
    def start_download_simple(self):
        """
        保险的单线程下载
        """
        for xyz_row in self.xyz_list:
            for xyz in xyz_row:
                xyz_dic={'x':xyz[0],'y':xyz[1],'z':xyz[2]}
                url=self.re_string%(xyz_dic)
                xyz_dic.update({'dir':self.dir_path,'prefix':self.prefix})
                filepath='%(dir)s/%(prefix)s_%(x)s_%(y)s_%(z)s.jpg'%(xyz_dic)
                #使用google代理来请求防止封ip
                import random
                g_min=1
                g_max=1000
                gint=random.randint(g_min,g_max)
                url='http://%d.gmodules.com/ig/proxy?url=%s'%(gint,urllib2.quote(url))
                self.down(url,filepath,xyz)
                
    def start_download_multh(self):
        """
        循环多线程下载图片
        请注意：容易被封ip
        """
        task_threads=[]
        for xyz_row in self.xyz_list:
            for xyz in xyz_row:
                xyz_dic={'x':xyz[0],'y':xyz[1],'z':xyz[2]}
                url=self.re_string%(xyz_dic)
                xyz_dic.update({'dir':self.dir_path,'prefix':self.prefix})
                filepath='%(dir)s/%(prefix)s_%(x)s_%(y)s_%(z)s.jpg'%(xyz_dic)
                #使用google代理来请求防止封ip
                import random
                g_min=1
                g_max=1000
                gint=random.randint(g_min,g_max)
                url='http://%d.gmodules.com/ig/proxy?url=%s'%(gint,urllib2.quote(url))
                t= threading.Thread( target=self.down,args=(url,filepath,xyz) )
                task_threads.append(t)
                t.start()
        for task in task_threads:
            task.join()
            
    def down(self,url,filepath,current_xyz,info={'scale':1,'corpbox':[]}):
        """
        down picture worker
        """
        try:
            print(">>>%s"%url)
            if self.cache_pics.has_key(url):
                jpg=self.cache_pics[url]
            else:
                web = urllib2.urlopen( url)
                jpg = web.read()
                
                response_type=web.info().getheader('Content-Type')
                if response_type.startswith('text/html'):
                    print '<<<error:%s'%str(web.info().getheader('X-HTTP-Status'))
                    #fix 404
                    if web.info().getheader('X-HTTP-Status')=='404':
                        info.update({'xyz':current_xyz})
                        import copy
                        temp_info=copy.deepcopy(info)
                        self.web_404_dic[filepath]=temp_info
                    return
                else:
                    #cache it
                    self.cache_pics[url]=jpg
                
            if self.web_404_dic.has_key(filepath):
                #this is fix down
                scale=info['scale']
                tmpfilepath='/tmp/%s'%(str(datetime.datetime.now()))
                File=open(tmpfilepath,'wb')
                File.write(jpg)
                File.close()
                im=Image.open(tmpfilepath)
                for idx in range(scale-1):
                    im_resized=im.resize((256*2,256*2), Image.ANTIALIAS)
                    box = info['corpbox'].pop(len(info['corpbox'])-1)
                    im=im_resized.crop(box)
                    
                print("<<<%s"%filepath)
                im.save(filepath)
                self.web_404_dic.pop(filepath)
                return
                
            print("<<<%s"%filepath)
            try:
                File = open( filepath,"wb" )
                File.write( jpg)
                File.close()
                return
            except IOError:
                print("!!!IO error")
                return
        except Exception,e:
            print("!!!error:%s"%str(e))
            return
            
    def getNewXYZ(self,xyz):
        #pdb.set_trace()
        x=xyz[0]
        y=xyz[1]
        z=xyz[2]
        
        z1=z-1
        x1=float((float(x)/float(2**z-1))*float(2**z1-1))
        y1=float((float(y)/float(2**z-1))*float(2**z1-1))
        if float(x1-int(x1))>=float(0.5):
            x1=int(x1)+1
        else:
            x1=int(x1)
        if float(y1-int(y1))>=float(0.5):
            y1=int(y1)+1
        else:
            y1=int(y1)
            
        print '%s->%s'%(str(xyz),(x1,y1,z1))
        return (x1,y1,z1)
        
    def fix_web_404(self):
        """
        尝试递归z_level来修复瓦片数据不存在情况
        """
        miss_count=len(self.web_404_dic)
        all_count=len(self.xyz_list)*len(self.xyz_list[0])
        if miss_count*2>=all_count:
            #miss too many that means z-level error
            raise Exception('太多404,请重新设置z_level到合适的value')
        
        for filepath in self.web_404_dic.keys():
            datadic=self.web_404_dic[filepath]
            xyz=datadic['xyz']
            datadic['scale']+=datadic['scale']
            #up
            new_xyz=self.getNewXYZ(xyz)
            #get corpbox
            #only get left-top
            temp_xyz=self.getNewXYZ((xyz[0]-1,xyz[1],xyz[2]))
            if temp_xyz[0]!=new_xyz[0]:
                left_be=True
            else:
                left_be=False
            
            temp_xyz=self.getNewXYZ((xyz[0],xyz[1]-1,xyz[2]))
            if temp_xyz[1]!=new_xyz[1]:
                top_be=True
            else:
                top_be=False
            
            if left_be and top_be:
                datadic['corpbox'].append((0,0,256,256))
            elif left_be and top_be is False:
                datadic['corpbox'].append((0,256,256,256+256))
            elif left_be is False and top_be:
                datadic['corpbox'].append((256,0,256+256,256))
            elif left_be is False and top_be is False:
                datadic['corpbox'].append((256,256,256+256,256+256))
                
            #new url
            xyz_dic={'x':new_xyz[0],'y':new_xyz[1],'z':new_xyz[2]}
            url=self.re_string%(xyz_dic)
            #使用google代理来请求防止封ip
            import random
            g_min=1
            g_max=1000
            gint=random.randint(g_min,g_max)
            url='http://%d.gmodules.com/ig/proxy?url=%s'%(gint,urllib2.quote(url))
            #down
            #pdb.set_trace()
            self.down(url,filepath,xyz,datadic)
            
class PictureGen():
    """
    拼接图片
    """
    def __init__(self,xyz_list,dir_path,out_filepath,prefix='gm',per_size=(256,256),force_out=False):
        """
        图片矩形描述列表
        文件夹位置
        """
        self.force_out=force_out
        self.per_size=per_size
        self.xyz_list=xyz_list
        self.dir_path=dir_path
        self.out_filepath=out_filepath
        self.prefix=prefix
        
        #计算输出图片大小
        len_w=len(self.xyz_list[0])
        len_h=len(self.xyz_list)
        
        self.out_size=(len_w*per_size[0],len_h*per_size[1])
        
    def gen(self):
        """
        取得左上的瓦片坐标开始拼接图片
        """
        lt=self.xyz_list[0][0]
        self.out_file=Image.new('RGBA', self.out_size)
        
        for xyz_row in self.xyz_list:
            for xyz in xyz_row:
                xyz_dic={'x':xyz[0],'y':xyz[1],'z':xyz[2]}
                xyz_dic.update({'dir':self.dir_path,'prefix':self.prefix})
                filepath='%(dir)s/%(prefix)s_%(x)s_%(y)s_%(z)s.jpg'%(xyz_dic)
                try:
                    fromImage = Image.open(filepath)
                except:
                    if self.force_out:
                        fromImage=Image.new('RGBA',self.per_size,(0,0,0))
                        draw=ImageDraw.Draw(fromImage)
                        draw.text( (0,0) ,str(xyz),(255,255,255))
                    else:
                        raise
                        
                self.out_file.paste(fromImage,( (xyz[0]-lt[0])*self.per_size[0],(xyz[1]-lt[1])*self.per_size[1]))
                
        self.out_file.save(self.out_filepath)
        
        
if __name__ == "__main__":
    #pdb.set_trace()
    #北京故宫经纬
    s=(39.921543,116.385174)
    e=(39.911399,116.396819)
    z=19
    dir_path='/home/smite/opt'
    import datetime
    out_filepath='/home/smite/opt/%s.jpg'%(str(datetime.datetime.now()))
    rect=RectCreater(s,e,z)
    res=ResourceGeter(rect.xyz_list,dir_path,'https://khms1.google.com/kh/v=128&src=app&x=%(x)s&y=%(y)s&z=%(z)s')
    counter=len(res.xyz_list)*len(res.xyz_list[0])
    print '共%d张图片等待下载。。。'%(counter)
    res.start_download_simple()
    counter=0
    #pdb.set_trace()
    while len(res.web_404_dic)!=0:
        counter+=1
        print '开始第%d次修复...'%counter
        print '需要修复%d张图片'%len(res.web_404_dic)
        res.fix_web_404()
        
    pic=PictureGen(rect.xyz_list,dir_path,out_filepath,force_out=True)
    pic.gen()